from typing import Generic, TypeVar, Optional, Generator, Any
import asyncio

_T = TypeVar('_T')


class Queue(Generic[_T]):
    def __init__(self, maxsize: int=0, *, loop: Optional[asyncio.AbstractEventLoop]=None) -> None: ...

    def close(self) -> None: ...

    async def wait_closed(self) -> None: ...

    @property
    def maxsize(self) -> int: ...

    @property
    def sync_q(self) -> _SyncQueueProxy[_T]: ...

    @property
    def async_q(self) -> _AsyncQueueProxy[_T]: ...


class _SyncQueueProxy(Generic[_T]):

    @property
    def maxsize(self) -> int: ...

    def task_done(self) -> None: ...

    def join(self) -> None: ...

    def qsize(self) -> int: ...

    def empty(self) -> bool: ...

    def full(self) -> bool: ...

    def put(self, item: _T, block: bool=True, timeout: Optional[float]=None) -> None: ...

    def get(self, block: bool=True, timeout: Optional[float]=None) -> _T: ...


    def put_nowait(self, item: _T) -> None: ...

    def get_nowait(self) -> _T: ...


class _AsyncQueueProxy(Generic[_T]):

    @property
    def maxsize(self) -> int: ...

    def task_done(self) -> None: ...

    async def join(self) -> None: ...

    def qsize(self) -> int: ...

    def empty(self) -> bool: ...

    def full(self) -> bool: ...

    async def put(self, item: _T) -> None: ...

    async def get(self) -> _T: ...


    def put_nowait(self, item: _T) -> None: ...

    def get_nowait(self) -> _T: ...


class PriorityQueue(Queue[_T], Generic[_T]): ...


class LifoQueue(Queue[_T], Generic[_T]): ...



import asyncio
import concurrent.futures
import unittest

from unittest import mock

import janus


class TestMixedMode(unittest.TestCase):
    def setUp(self):
        self.loop = asyncio.new_event_loop()
        self.executor = concurrent.futures.ThreadPoolExecutor(max_workers=5)
        self.loop.set_default_executor(self.executor)
        asyncio.set_event_loop(None)

    def tearDown(self):
        self.executor.shutdown()
        self.loop.close()

    def test_ctor_loop(self):
        loop = mock.Mock()
        q = janus.Queue(loop=loop)
        self.assertIs(q._loop, loop)

        q = janus.Queue(loop=self.loop)
        self.assertIs(q._loop, self.loop)

    def test_ctor_noloop(self):
        asyncio.set_event_loop(self.loop)
        q = janus.Queue()
        self.assertIs(q._loop, self.loop)

    def test_maxsize(self):
        q = janus.Queue(5, loop=self.loop)
        self.assertIs(5, q.maxsize)

    def test_maxsize_named_param(self):
        q = janus.Queue(maxsize=7, loop=self.loop)
        self.assertIs(7, q.maxsize)

    def test_maxsize_default(self):
        q = janus.Queue(loop=self.loop)
        self.assertIs(0, q.maxsize)

    def test_unfinished(self):
        q = janus.Queue(loop=self.loop)
        self.assertEqual(q.sync_q.unfinished_tasks, 0)
        self.assertEqual(q.async_q.unfinished_tasks, 0)
        q.sync_q.put(1)
        self.assertEqual(q.sync_q.unfinished_tasks, 1)
        self.assertEqual(q.async_q.unfinished_tasks, 1)
        q.sync_q.get()
        self.assertEqual(q.sync_q.unfinished_tasks, 1)
        self.assertEqual(q.async_q.unfinished_tasks, 1)
        q.sync_q.task_done()
        self.assertEqual(q.sync_q.unfinished_tasks, 0)
        self.assertEqual(q.async_q.unfinished_tasks, 0)

    def test_sync_put_async_get(self):
        q = janus.Queue(loop=self.loop)

        def threaded():
            for i in range(5):
                q.sync_q.put(i)

        async def go():
            f = self.loop.run_in_executor(None, threaded)
            for i in range(5):
                val = await q.async_q.get()
                self.assertEqual(val, i)

            self.assertTrue(q.async_q.empty())

            await f

        for i in range(3):
            self.loop.run_until_complete(go())

    def test_sync_put_async_join(self):
        q = janus.Queue(loop=self.loop)

        for i in range(5):
            q.sync_q.put(i)

        async def do_work():
            await asyncio.sleep(1)
            while True:
                await q.async_q.get()
                q.async_q.task_done()

        task = self.loop.create_task(do_work())

        async def wait_for_empty_queue():
            await q.async_q.join()
            task.cancel()

        self.loop.run_until_complete(wait_for_empty_queue())

    def test_async_put_sync_get(self):
        q = janus.Queue(loop=self.loop)

        def threaded():
            for i in range(5):
                val = q.sync_q.get()
                self.assertEqual(val, i)

        async def go():
            f = self.loop.run_in_executor(None, threaded)
            for i in range(5):
                await q.async_q.put(i)

            await f
            self.assertTrue(q.async_q.empty())

        for i in range(3):
            self.loop.run_until_complete(go())

    def test_sync_join_async_done(self):
        q = janus.Queue(loop=self.loop)

        def threaded():
            for i in range(5):
                q.sync_q.put(i)
            q.sync_q.join()

        async def go():
            f = self.loop.run_in_executor(None, threaded)
            for i in range(5):
                val = await q.async_q.get()
                self.assertEqual(val, i)
                q.async_q.task_done()

            self.assertTrue(q.async_q.empty())

            await f

        for i in range(3):
            self.loop.run_until_complete(go())

    def test_async_join_async_done(self):
        q = janus.Queue(loop=self.loop)

        def threaded():
            for i in range(5):
                val = q.sync_q.get()
                self.assertEqual(val, i)
                q.sync_q.task_done()

        async def go():
            f = self.loop.run_in_executor(None, threaded)
            for i in range(5):
                await q.async_q.put(i)

            await q.async_q.join()

            await f
            self.assertTrue(q.async_q.empty())

        for i in range(3):
            self.loop.run_until_complete(go())

    def test_wait_without_closing(self):
        q = janus.Queue(loop=self.loop)

        with self.assertRaises(RuntimeError):
            self.loop.run_until_complete(q.wait_closed())

        q.close()
        self.loop.run_until_complete(q.wait_closed())

    def test_modifying_forbidden_after_closing(self):
        q = janus.Queue(loop=self.loop)
        q.close()

        with self.assertRaises(RuntimeError):
            q.sync_q.put(5)

        with self.assertRaises(RuntimeError):
            q.sync_q.get()

        with self.assertRaises(RuntimeError):
            q.sync_q.task_done()

        with self.assertRaises(RuntimeError):
            self.loop.run_until_complete(q.async_q.put(5))

        with self.assertRaises(RuntimeError):
            q.async_q.put_nowait(5)

        with self.assertRaises(RuntimeError):
            q.async_q.get_nowait()

        with self.assertRaises(RuntimeError):
            self.loop.run_until_complete(q.sync_q.task_done())

        self.loop.run_until_complete(q.wait_closed())

    def test_double_closing(self):
        q = janus.Queue(loop=self.loop)
        q.close()
        q.close()
        self.loop.run_until_complete(q.wait_closed())

    def test_closed(self):
        q = janus.Queue(loop=self.loop)
        self.assertFalse(q.closed)
        self.assertFalse(q.async_q.closed)
        self.assertFalse(q.sync_q.closed)
        q.close()
        self.assertTrue(q.closed)
        self.assertTrue(q.async_q.closed)
        self.assertTrue(q.sync_q.closed)



import asyncio
import logging
import threading
from asyncio import QueueEmpty as AsyncQueueEmpty
from asyncio import QueueFull as AsyncQueueFull
from collections import deque
from heapq import heappop, heappush
from queue import Empty as SyncQueueEmpty
from queue import Full as SyncQueueFull

__version__ = '0.4.0'

log = logging.getLogger(__package__)


try:
    ensure_future = asyncio.ensure_future
except AttributeError:
    ensure_future = getattr(asyncio, 'async')


class Queue:
    def __init__(self, maxsize=0, *, loop=None):
        if loop is None:
            loop = asyncio.get_event_loop()

        self._loop = loop
        self._maxsize = maxsize

        self._init(maxsize)

        self._unfinished_tasks = 0

        self._sync_mutex = threading.Lock()
        self._sync_not_empty = threading.Condition(self._sync_mutex)
        self._sync_not_full = threading.Condition(self._sync_mutex)
        self._all_tasks_done = threading.Condition(self._sync_mutex)

        self._async_mutex = asyncio.Lock(loop=loop)
        self._async_not_empty = asyncio.Condition(self._async_mutex, loop=loop)
        self._async_not_full = asyncio.Condition(self._async_mutex, loop=loop)
        self._finished = asyncio.Event(loop=self._loop)
        self._finished.set()

        self._closing = False
        self._pending = set()

        def checked_call_soon_threadsafe(callback, *args):
            try:
                loop.call_soon_threadsafe(callback, *args)
            except RuntimeError:
                # swallowing agreed in #2
                pass
        self._call_soon_threadsafe = checked_call_soon_threadsafe

        def checked_call_soon(callback, *args):
            if not loop.is_closed():
                loop.call_soon(callback, *args)
        self._call_soon = checked_call_soon

        self._sync_queue = _SyncQueueProxy(self)
        self._async_queue = _AsyncQueueProxy(self)

    def close(self):
        with self._sync_mutex:
            self._closing = True
            for fut in self._pending:
                fut.cancel()

    async def wait_closed(self):
        # should be called from loop after close().
        # Nobody should put/get at this point,
        # so lock acquiring is not required
        if not self._closing:
            raise RuntimeError("Waiting for non-closed queue")
        if not self._pending:
            return
        await asyncio.wait(self._pending, loop=self._loop)

    @property
    def closed(self):
        return self._closing and not self._pending

    @property
    def maxsize(self):
        return self._maxsize

    @property
    def sync_q(self):
        return self._sync_queue

    @property
    def async_q(self):
        return self._async_queue

    # Override these methods to implement other queue organizations
    # (e.g. stack or priority queue).
    # These will only be called with appropriate locks held

    def _init(self, maxsize):
        self._queue = deque()

    def _qsize(self):
        return len(self._queue)

    # Put a new item in the queue
    def _put(self, item):
        self._queue.append(item)

    # Get an item from the queue
    def _get(self):
        return self._queue.popleft()

    def _put_internal(self, item):
        self._put(item)
        self._unfinished_tasks += 1
        self._finished.clear()

    def _notify_sync_not_empty(self):
        def f():
            with self._sync_mutex:
                self._sync_not_empty.notify()

        self._loop.run_in_executor(None, f)

    def _notify_sync_not_full(self):
        def f():
            with self._sync_mutex:
                self._sync_not_full.notify()

        fut = self._loop.run_in_executor(None, f)
        fut.add_done_callback(self._pending.discard)
        self._pending.add(fut)

    def _notify_async_not_empty(self, *, threadsafe):
        async def f():
            async with self._async_mutex:
                self._async_not_empty.notify()

        def task_maker():
            task = ensure_future(f(), loop=self._loop)
            task.add_done_callback(self._pending.discard)
            self._pending.add(task)

        if threadsafe:
            self._call_soon_threadsafe(task_maker)
        else:
            self._call_soon(task_maker)

    def _notify_async_not_full(self, *, threadsafe):
        async def f():
            async with self._async_mutex:
                self._async_not_full.notify()

        def task_maker():
            task = ensure_future(f(), loop=self._loop)
            task.add_done_callback(self._pending.discard)
            self._pending.add(task)

        if threadsafe:
            self._call_soon_threadsafe(task_maker)
        else:
            self._call_soon(task_maker)

    def _check_closing(self):
        if self._closing:
            raise RuntimeError('Modification of closed queue is forbidden')


class _SyncQueueProxy:
    '''Create a queue object with a given maximum size.

    If maxsize is <= 0, the queue size is infinite.
    '''

    def __init__(self, parent):
        self._parent = parent

    @property
    def maxsize(self):
        return self._parent._maxsize

    @property
    def closed(self):
        return self._parent.closed

    def task_done(self):
        '''Indicate that a formerly enqueued task is complete.

        Used by Queue consumer threads.  For each get() used to fetch a task,
        a subsequent call to task_done() tells the queue that the processing
        on the task is complete.

        If a join() is currently blocking, it will resume when all items
        have been processed (meaning that a task_done() call was received
        for every item that had been put() into the queue).

        Raises a ValueError if called more times than there were items
        placed in the queue.
        '''
        self._parent._check_closing()
        with self._parent._all_tasks_done:
            unfinished = self._parent._unfinished_tasks - 1
            if unfinished <= 0:
                if unfinished < 0:
                    raise ValueError('task_done() called too many times')
                self._parent._all_tasks_done.notify_all()
                self._parent._loop.call_soon_threadsafe(
                    self._parent._finished.set)
            self._parent._unfinished_tasks = unfinished

    def join(self):
        '''Blocks until all items in the Queue have been gotten and processed.

        The count of unfinished tasks goes up whenever an item is added to the
        queue. The count goes down whenever a consumer thread calls task_done()
        to indicate the item was retrieved and all work on it is complete.

        When the count of unfinished tasks drops to zero, join() unblocks.
        '''
        with self._parent._all_tasks_done:
            while self._parent._unfinished_tasks:
                self._parent._all_tasks_done.wait()

    def qsize(self):
        '''Return the approximate size of the queue (not reliable!).'''
        return self._parent._qsize()

    @property
    def unfinished_tasks(self):
        '''Return the number of unfinished tasks.'''
        return self._parent._unfinished_tasks

    def empty(self):
        '''Return True if the queue is empty, False otherwise (not reliable!).

        This method is likely to be removed at some point.  Use qsize() == 0
        as a direct substitute, but be aware that either approach risks a race
        condition where a queue can grow before the result of empty() or
        qsize() can be used.

        To create code that needs to wait for all queued tasks to be
        completed, the preferred technique is to use the join() method.
        '''
        return not self._parent._qsize()

    def full(self):
        '''Return True if the queue is full, False otherwise (not reliable!).

        This method is likely to be removed at some point.  Use qsize() >= n
        as a direct substitute, but be aware that either approach risks a race
        condition where a queue can shrink before the result of full() or
        qsize() can be used.
        '''
        return 0 < self._parent._maxsize <= self._parent._qsize()

    def put(self, item, block=True, timeout=None):
        '''Put an item into the queue.

        If optional args 'block' is true and 'timeout' is None (the default),
        block if necessary until a free slot is available. If 'timeout' is
        a non-negative number, it blocks at most 'timeout' seconds and raises
        the Full exception if no free slot was available within that time.
        Otherwise ('block' is false), put an item on the queue if a free slot
        is immediately available, else raise the Full exception ('timeout'
        is ignored in that case).
        '''
        self._parent._check_closing()
        with self._parent._sync_not_full:
            if self._parent._maxsize > 0:
                if not block:
                    if self._parent._qsize() >= self._parent._maxsize:
                        raise SyncQueueFull
                elif timeout is None:
                    while self._parent._qsize() >= self._parent._maxsize:
                        self._parent._sync_not_full.wait()
                elif timeout < 0:
                    raise ValueError("'timeout' must be a non-negative number")
                else:
                    time = self._parent._loop.time
                    endtime = time() + timeout
                    while self._parent._qsize() >= self._parent._maxsize:
                        remaining = endtime - time()
                        if remaining <= 0.0:
                            raise SyncQueueFull
                        self._parent._sync_not_full.wait(remaining)
            self._parent._put_internal(item)
            self._parent._sync_not_empty.notify()
            self._parent._notify_async_not_empty(threadsafe=True)

    def get(self, block=True, timeout=None):
        '''Remove and return an item from the queue.

        If optional args 'block' is true and 'timeout' is None (the default),
        block if necessary until an item is available. If 'timeout' is
        a non-negative number, it blocks at most 'timeout' seconds and raises
        the Empty exception if no item was available within that time.
        Otherwise ('block' is false), return an item if one is immediately
        available, else raise the Empty exception ('timeout' is ignored
        in that case).
        '''
        self._parent._check_closing()
        with self._parent._sync_not_empty:
            if not block:
                if not self._parent._qsize():
                    raise SyncQueueEmpty
            elif timeout is None:
                while not self._parent._qsize():
                    self._parent._sync_not_empty.wait()
            elif timeout < 0:
                raise ValueError("'timeout' must be a non-negative number")
            else:
                time = self._parent._loop.time
                endtime = time() + timeout
                while not self._parent._qsize():
                    remaining = endtime - time()
                    if remaining <= 0.0:
                        raise SyncQueueEmpty
                    self._parent._sync_not_empty.wait(remaining)
            item = self._parent._get()
            self._parent._sync_not_full.notify()
            self._parent._notify_async_not_full(threadsafe=True)
            return item

    def put_nowait(self, item):
        '''Put an item into the queue without blocking.

        Only enqueue the item if a free slot is immediately available.
        Otherwise raise the Full exception.
        '''
        return self.put(item, block=False)

    def get_nowait(self):
        '''Remove and return an item from the queue without blocking.

        Only get an item if one is immediately available. Otherwise
        raise the Empty exception.
        '''
        return self.get(block=False)


class _AsyncQueueProxy:
    '''Create a queue object with a given maximum size.

    If maxsize is <= 0, the queue size is infinite.
    '''

    def __init__(self, parent):
        self._parent = parent

    @property
    def closed(self):
        return self._parent.closed

    def qsize(self):
        """Number of items in the queue."""
        return self._parent._qsize()

    @property
    def unfinished_tasks(self):
        '''Return the number of unfinished tasks.'''
        return self._parent._unfinished_tasks

    @property
    def maxsize(self):
        """Number of items allowed in the queue."""
        return self._parent._maxsize

    def empty(self):
        """Return True if the queue is empty, False otherwise."""
        return self.qsize() == 0

    def full(self):
        """Return True if there are maxsize items in the queue.

        Note: if the Queue was initialized with maxsize=0 (the default),
        then full() is never True.
        """
        if self._parent._maxsize <= 0:
            return False
        else:
            return self.qsize() >= self._parent._maxsize

    async def put(self, item):
        """Put an item into the queue.

        Put an item into the queue. If the queue is full, wait until a free
        slot is available before adding item.

        This method is a coroutine.
        """
        self._parent._check_closing()
        async with self._parent._async_not_full:
            self._parent._sync_mutex.acquire()
            locked = True
            try:
                if self._parent._maxsize > 0:
                    do_wait = True
                    while do_wait:
                        do_wait = (
                            self._parent._qsize() >= self._parent._maxsize
                        )
                        if do_wait:
                            locked = False
                            self._parent._sync_mutex.release()
                            await self._parent._async_not_full.wait()
                            self._parent._sync_mutex.acquire()
                            locked = True

                self._parent._put_internal(item)
                self._parent._async_not_empty.notify()
                self._parent._notify_sync_not_empty()
            finally:
                if locked:
                    self._parent._sync_mutex.release()

    def put_nowait(self, item):
        """Put an item into the queue without blocking.

        If no free slot is immediately available, raise QueueFull.
        """
        self._parent._check_closing()
        with self._parent._sync_mutex:
            if self._parent._maxsize > 0:
                if self._parent._qsize() >= self._parent._maxsize:
                    raise AsyncQueueFull

            self._parent._put_internal(item)
            self._parent._notify_async_not_empty(threadsafe=False)
            self._parent._notify_sync_not_empty()

    async def get(self):
        """Remove and return an item from the queue.

        If queue is empty, wait until an item is available.

        This method is a coroutine.
        """
        self._parent._check_closing()
        async with self._parent._async_not_empty:
            self._parent._sync_mutex.acquire()
            locked = True
            try:
                do_wait = True
                while do_wait:
                    do_wait = self._parent._qsize() == 0

                    if do_wait:
                        locked = False
                        self._parent._sync_mutex.release()
                        await self._parent._async_not_empty.wait()
                        self._parent._sync_mutex.acquire()
                        locked = True

                item = self._parent._get()
                self._parent._async_not_full.notify()
                self._parent._notify_sync_not_full()
                return item
            finally:
                if locked:
                    self._parent._sync_mutex.release()

    def get_nowait(self):
        """Remove and return an item from the queue.

        Return an item if one is immediately available, else raise QueueEmpty.
        """
        self._parent._check_closing()
        with self._parent._sync_mutex:
            if self._parent._qsize() == 0:
                raise AsyncQueueEmpty

            item = self._parent._get()
            self._parent._notify_async_not_full(threadsafe=False)
            self._parent._notify_sync_not_full()
            return item

    def task_done(self):
        """Indicate that a formerly enqueued task is complete.

        Used by queue consumers. For each get() used to fetch a task,
        a subsequent call to task_done() tells the queue that the processing
        on the task is complete.

        If a join() is currently blocking, it will resume when all items have
        been processed (meaning that a task_done() call was received for every
        item that had been put() into the queue).

        Raises ValueError if called more times than there were items placed in
        the queue.
        """
        self._parent._check_closing()
        with self._parent._all_tasks_done:
            if self._parent._unfinished_tasks <= 0:
                raise ValueError('task_done() called too many times')
            self._parent._unfinished_tasks -= 1
            if self._parent._unfinished_tasks == 0:
                self._parent._finished.set()
                self._parent._all_tasks_done.notify_all()

    async def join(self):
        """Block until all items in the queue have been gotten and processed.

        The count of unfinished tasks goes up whenever an item is added to the
        queue. The count goes down whenever a consumer calls task_done() to
        indicate that the item was retrieved and all work on it is complete.
        When the count of unfinished tasks drops to zero, join() unblocks.
        """
        while True:
            with self._parent._sync_mutex:
                if self._parent._unfinished_tasks == 0:
                    break
            await self._parent._finished.wait()


class PriorityQueue(Queue):
    '''Variant of Queue that retrieves open entries in priority order
    (lowest first).

    Entries are typically tuples of the form:  (priority number, data).

    '''

    def _init(self, maxsize):
        self._queue = []

    def _qsize(self):
        return len(self._queue)

    def _put(self, item):
        heappush(self._queue, item)

    def _get(self):
        return heappop(self._queue)


class LifoQueue(Queue):
    '''Variant of Queue that retrieves most recently added entries first.'''

    def _init(self, maxsize):
        self._queue = deque()

    def _qsize(self):
        return len(self._queue)

    def _put(self, item):
        self._queue.append(item)

    def _get(self):
        return self._queue.pop()



import codecs
from setuptools import setup, find_packages
import os
import re
import sys

from setuptools.command.test import test as TestCommand

PY_33 = sys.version_info < (3, 4)
PY_35 = sys.version_info >= (3, 5)


class PyTest(TestCommand):
    user_options = [('pytest-args=', 'a', "Arguments to pass to py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = []

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main(self.pytest_args)
        sys.exit(errno)


with codecs.open(os.path.join(os.path.abspath(os.path.dirname(
        __file__)), 'janus', '__init__.py'), 'r', 'latin1') as fp:
    try:
        version = re.findall(r"^__version__ = '([^']+)'$", fp.read(), re.M)[0]
    except IndexError:
        raise RuntimeError('Unable to determine version.')


def read(f):
    return open(os.path.join(os.path.dirname(__file__), f)).read().strip()


install_requires = []

if PY_33:
    install_requires.append('asyncio')

# if not PY_35:
#     install_requires.append('typing')

tests_require = install_requires + ['pytest']
extras_require = {}


setup(
    name='janus',
    version=version,
    description=("Mixed sync-async queue to interoperate between "
                 "asyncio tasks and classic threads"),
    long_description='\n\n'.join((read('README.rst'), read('CHANGES.rst'))),
    classifiers=[
        'License :: OSI Approved :: Apache Software License',
        'Intended Audience :: Developers', 'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Software Development :: Libraries',
        'Framework :: AsyncIO',
    ],
    author='Andrew Svetlov',
    author_email='andrew.svetlov@gmail.com',
    url='https://github.com/aio-libs/janus/',
    license='Apache 2',
    packages=find_packages(),
    python_requires='>=3.5.3',
    install_requires=install_requires,
    tests_require=tests_require,
    cmdclass={'test': PyTest},
    include_package_data=True,
    extras_require=extras_require)



"""Tests for queues.py"""

import asyncio
import concurrent.futures
import unittest

import janus


class _QueueTestBase(unittest.TestCase):
    def setUp(self):
        self.loop = asyncio.new_event_loop()
        self.executor = concurrent.futures.ThreadPoolExecutor(max_workers=5)
        self.loop.set_default_executor(self.executor)
        asyncio.set_event_loop(None)

    def tearDown(self):
        self.executor.shutdown()
        self.loop.close()


class QueueBasicTests(_QueueTestBase):
    def _test_repr_or_str(self, fn, expect_id):
        """Test Queue's repr or str.

        fn is repr or str. expect_id is True if we expect the Queue's id to
        appear in fn(Queue()).
        """

        _q = janus.Queue(loop=self.loop)
        q = _q.async_q
        self.assertTrue(fn(q).startswith('<Queue'), fn(q))
        id_is_present = hex(id(q)) in fn(q)
        self.assertEqual(expect_id, id_is_present)

        async def add_getter():
            _q = janus.Queue(loop=self.loop)
            q = _q.async_q
            # Start a task that waits to get.
            self.loop.create_task(q.get())
            # Let it start waiting.
            await asyncio.sleep(0.1)
            self.assertTrue('_getters[1]' in fn(q))
            # resume q.get coroutine to finish generator
            q.put_nowait(0)

        self.loop.run_until_complete(add_getter())

        async def add_putter():
            _q = janus.Queue(maxsize=1, loop=self.loop)
            q = _q.async_q
            q.put_nowait(1)
            # Start a task that waits to put.
            self.loop.create_task(q.put(2))
            # Let it start waiting.
            await asyncio.sleep(0.1)
            self.assertTrue('_putters[1]' in fn(q))
            # resume q.put coroutine to finish generator
            q.get_nowait()

        self.loop.run_until_complete(add_putter())

        _q = janus.Queue(loop=self.loop)
        q = _q.async_q
        q.put_nowait(1)
        self.assertTrue('_queue=[1]' in fn(q))

    # def test_repr(self):
    #     self._test_repr_or_str(repr, True)

    # def test_str(self):
    #     self._test_repr_or_str(str, False)

    def test_empty(self):
        _q = janus.Queue(loop=self.loop)
        q = _q.async_q
        self.assertTrue(q.empty())
        q.put_nowait(1)
        self.assertFalse(q.empty())
        self.assertEqual(1, q.get_nowait())
        self.assertTrue(q.empty())

        self.assertFalse(_q._sync_mutex.locked())
        _q.close()
        self.loop.run_until_complete(_q.wait_closed())

    def test_full(self):
        _q = janus.Queue(loop=self.loop)
        q = _q.async_q
        self.assertFalse(q.full())

        _q = janus.Queue(maxsize=1, loop=self.loop)
        q = _q.async_q
        q.put_nowait(1)
        self.assertTrue(q.full())

        self.assertFalse(_q._sync_mutex.locked())
        _q.close()
        self.loop.run_until_complete(_q.wait_closed())

    def test_order(self):
        _q = janus.Queue(loop=self.loop)
        q = _q.async_q
        for i in [1, 3, 2]:
            q.put_nowait(i)

        items = [q.get_nowait() for _ in range(3)]
        self.assertEqual([1, 3, 2], items)

        self.assertFalse(_q._sync_mutex.locked())
        _q.close()
        self.loop.run_until_complete(_q.wait_closed())

    def test_maxsize(self):
        _q = janus.Queue(maxsize=2, loop=self.loop)
        q = _q.async_q
        self.assertEqual(2, q.maxsize)
        have_been_put = []

        fut = self.loop.create_future()

        async def putter():
            for i in range(3):
                await q.put(i)
                have_been_put.append(i)
                if i == q.maxsize - 1:
                    fut.set_result(None)
            return True

        async def test():
            t = self.loop.create_task(putter())
            await fut

            # The putter is blocked after putting two items.
            self.assertEqual([0, 1], have_been_put)
            self.assertEqual(0, q.get_nowait())

            # Let the putter resume and put last item.
            await t
            self.assertEqual([0, 1, 2], have_been_put)
            self.assertEqual(1, q.get_nowait())
            self.assertEqual(2, q.get_nowait())

        self.loop.run_until_complete(test())

        self.assertFalse(_q._sync_mutex.locked())
        _q.close()
        self.loop.run_until_complete(_q.wait_closed())


class QueueGetTests(_QueueTestBase):
    def test_blocking_get(self):
        _q = janus.Queue(loop=self.loop)
        q = _q.async_q
        q.put_nowait(1)

        async def queue_get():
            return (await q.get())

        res = self.loop.run_until_complete(queue_get())
        self.assertEqual(1, res)

        self.assertFalse(_q._sync_mutex.locked())
        _q.close()
        self.loop.run_until_complete(_q.wait_closed())

    def test_get_with_putters(self):
        _q = janus.Queue(1, loop=self.loop)
        q = _q.async_q
        q.put_nowait(1)

        fut = self.loop.create_future()

        async def put():
            t = janus.ensure_future(q.put(2), loop=self.loop)
            await asyncio.sleep(0.01)
            fut.set_result(None)
            return t

        t = self.loop.run_until_complete(put())

        res = self.loop.run_until_complete(q.get())
        self.assertEqual(1, res)

        self.loop.run_until_complete(t)
        self.assertEqual(1, q.qsize())

        self.assertFalse(_q._sync_mutex.locked())
        _q.close()
        self.loop.run_until_complete(_q.wait_closed())

    def test_blocking_get_wait(self):
        _q = janus.Queue(loop=self.loop)
        q = _q.async_q
        started = asyncio.Event(loop=self.loop)
        finished = False

        async def queue_get():
            nonlocal finished
            started.set()
            res = await q.get()
            finished = True
            return res

        async def queue_put():
            self.loop.call_later(0.01, q.put_nowait, 1)
            queue_get_task = self.loop.create_task(queue_get())
            await started.wait()
            self.assertFalse(finished)
            res = await queue_get_task
            self.assertTrue(finished)
            return res

        res = self.loop.run_until_complete(queue_put())
        self.assertEqual(1, res)

        self.assertFalse(_q._sync_mutex.locked())
        _q.close()
        self.loop.run_until_complete(_q.wait_closed())

    def test_nonblocking_get(self):
        _q = janus.Queue(loop=self.loop)
        q = _q.async_q
        q.put_nowait(1)
        self.assertEqual(1, q.get_nowait())

        _q.close()
        self.loop.run_until_complete(_q.wait_closed())

    def test_nonblocking_get_exception(self):
        _q = janus.Queue(loop=self.loop)
        self.assertRaises(asyncio.QueueEmpty, _q.async_q.get_nowait)

        self.assertFalse(_q._sync_mutex.locked())
        _q.close()
        self.loop.run_until_complete(_q.wait_closed())

    def test_get_cancelled(self):
        _q = janus.Queue(loop=self.loop)
        q = _q.async_q

        async def queue_get():
            return await asyncio.wait_for(q.get(), 0.051)

        async def test():
            get_task = self.loop.create_task(queue_get())
            await asyncio.sleep(0.01)  # let the task start
            q.put_nowait(1)
            return await get_task

        self.assertEqual(1, self.loop.run_until_complete(test()))

        self.assertFalse(_q._sync_mutex.locked())
        _q.close()
        self.loop.run_until_complete(_q.wait_closed())

    def test_get_cancelled_race(self):
        _q = janus.Queue(loop=self.loop)
        q = _q.async_q

        f1 = self.loop.create_future()

        async def g1():
            f1.set_result(None)
            await q.get()

        t1 = self.loop.create_task(g1())
        t2 = self.loop.create_task(q.get())

        self.loop.run_until_complete(f1)
        self.loop.run_until_complete(asyncio.sleep(0.01))
        t1.cancel()

        with self.assertRaises(asyncio.CancelledError):
            self.loop.run_until_complete(t1)
        self.assertTrue(t1.done())
        q.put_nowait('a')

        self.loop.run_until_complete(t2)
        self.assertEqual(t2.result(), 'a')

        self.assertFalse(_q._sync_mutex.locked())
        _q.close()
        self.loop.run_until_complete(_q.wait_closed())

    def test_get_with_waiting_putters(self):
        _q = janus.Queue(loop=self.loop, maxsize=1)
        q = _q.async_q

        self.loop.create_task(q.put('a'))
        self.loop.create_task(q.put('b'))

        self.loop.run_until_complete(asyncio.sleep(0.01, loop=self.loop))

        self.assertEqual(self.loop.run_until_complete(q.get()), 'a')
        self.assertEqual(self.loop.run_until_complete(q.get()), 'b')

        self.assertFalse(_q._sync_mutex.locked())
        _q.close()
        self.loop.run_until_complete(_q.wait_closed())


class QueuePutTests(_QueueTestBase):
    def test_blocking_put(self):
        _q = janus.Queue(loop=self.loop)
        q = _q.async_q

        async def queue_put():
            # No maxsize, won't block.
            await q.put(1)

        self.loop.run_until_complete(queue_put())

        self.assertFalse(_q._sync_mutex.locked())
        _q.close()
        self.loop.run_until_complete(_q.wait_closed())

    def test_blocking_put_wait(self):
        _q = janus.Queue(maxsize=1, loop=self.loop)
        q = _q.async_q
        started = asyncio.Event(loop=self.loop)
        finished = False

        async def queue_put():
            nonlocal finished
            started.set()
            await q.put(1)
            await q.put(2)
            finished = True

        async def queue_get():
            self.loop.call_later(0.01, q.get_nowait)
            queue_put_task = self.loop.create_task(queue_put())
            await started.wait()
            self.assertFalse(finished)
            await queue_put_task
            self.assertTrue(finished)

        self.loop.run_until_complete(queue_get())

        self.assertFalse(_q._sync_mutex.locked())
        _q.close()
        self.loop.run_until_complete(_q.wait_closed())

    def test_nonblocking_put(self):
        _q = janus.Queue(loop=self.loop)
        q = _q.async_q
        q.put_nowait(1)
        self.assertEqual(1, q.get_nowait())

        self.assertFalse(_q._sync_mutex.locked())
        _q.close()
        self.loop.run_until_complete(_q.wait_closed())

    def test_nonblocking_put_exception(self):
        _q = janus.Queue(maxsize=1, loop=self.loop)
        q = _q.async_q
        q.put_nowait(1)
        self.assertRaises(asyncio.QueueFull, q.put_nowait, 2)

        self.assertFalse(_q._sync_mutex.locked())
        _q.close()
        self.loop.run_until_complete(_q.wait_closed())

    def test_float_maxsize(self):
        _q = janus.Queue(maxsize=1.3, loop=self.loop)
        q = _q.async_q
        q.put_nowait(1)
        q.put_nowait(2)
        self.assertTrue(q.full())
        self.assertRaises(asyncio.QueueFull, q.put_nowait, 3)

        _q.close()
        self.loop.run_until_complete(_q.wait_closed())

        _q = janus.Queue(maxsize=1.3, loop=self.loop)
        q = _q.async_q

        async def queue_put():
            await q.put(1)
            await q.put(2)
            self.assertTrue(q.full())

        self.loop.run_until_complete(queue_put())

        self.assertFalse(_q._sync_mutex.locked())
        _q.close()
        self.loop.run_until_complete(_q.wait_closed())

    def test_put_cancelled(self):
        _q = janus.Queue(loop=self.loop)
        q = _q.async_q

        async def queue_put():
            await q.put(1)
            return True

        async def test():
            return (await q.get())

        t = self.loop.create_task(queue_put())
        self.assertEqual(1, self.loop.run_until_complete(test()))
        self.assertTrue(t.done())
        self.assertTrue(t.result())

        self.assertFalse(_q._sync_mutex.locked())
        _q.close()
        self.loop.run_until_complete(_q.wait_closed())

    def test_put_cancelled_race(self):
        _q = janus.Queue(loop=self.loop, maxsize=1)
        q = _q.async_q

        put_a = self.loop.create_task(q.put('a'))
        put_b = self.loop.create_task(q.put('b'))
        put_c = self.loop.create_task(q.put('X'))

        self.loop.run_until_complete(put_a)
        self.assertFalse(put_b.done())

        put_c.cancel()

        with self.assertRaises(asyncio.CancelledError):
            self.loop.run_until_complete(put_c)

        async def go():
            a = await q.get()
            self.assertEqual(a, 'a')
            b = await q.get()
            self.assertEqual(b, 'b')
            self.assertTrue(put_b.done())

        self.loop.run_until_complete(go())

        self.assertFalse(_q._sync_mutex.locked())
        _q.close()
        self.loop.run_until_complete(_q.wait_closed())

    def test_put_with_waiting_getters(self):
        fut = self.loop.create_future()

        async def go():
            fut.set_result(None)
            ret = await q.get()
            return ret

        async def put():
            await q.put('a')

        _q = janus.Queue(loop=self.loop)
        q = _q.async_q
        t = self.loop.create_task(go())
        self.loop.run_until_complete(fut)
        self.loop.run_until_complete(put())
        self.assertEqual(self.loop.run_until_complete(t), 'a')

        self.assertFalse(_q._sync_mutex.locked())
        _q.close()
        self.loop.run_until_complete(_q.wait_closed())


class LifoQueueTests(_QueueTestBase):
    def test_order(self):
        _q = janus.LifoQueue(loop=self.loop)
        q = _q.async_q
        for i in [1, 3, 2]:
            q.put_nowait(i)

        items = [q.get_nowait() for _ in range(3)]
        self.assertEqual([2, 3, 1], items)

        self.assertFalse(_q._sync_mutex.locked())
        _q.close()
        self.loop.run_until_complete(_q.wait_closed())


class PriorityQueueTests(_QueueTestBase):
    def test_order(self):
        _q = janus.PriorityQueue(loop=self.loop)
        q = _q.async_q
        for i in [1, 3, 2]:
            q.put_nowait(i)

        items = [q.get_nowait() for _ in range(3)]
        self.assertEqual([1, 2, 3], items)

        self.assertFalse(_q._sync_mutex.locked())
        _q.close()
        self.loop.run_until_complete(_q.wait_closed())


class _QueueJoinTestMixin:

    q_class = None

    def test_task_done_underflow(self):
        _q = self.q_class(loop=self.loop)
        q = _q.async_q
        self.assertRaises(ValueError, q.task_done)

        self.assertFalse(_q._sync_mutex.locked())
        _q.close()
        self.loop.run_until_complete(_q.wait_closed())

    def test_task_done(self):
        _q = self.q_class(loop=self.loop)
        q = _q.async_q
        for i in range(100):
            q.put_nowait(i)

        accumulator = 0

        # Two workers get items from the queue and call task_done after each.
        # Join the queue and assert all items have been processed.
        running = True

        async def worker():
            nonlocal accumulator

            while running:
                item = await q.get()
                accumulator += item
                q.task_done()

        async def test():
            tasks = [self.loop.create_task(worker())
                     for index in range(2)]

            await q.join()
            return tasks

        tasks = self.loop.run_until_complete(test())
        self.assertEqual(sum(range(100)), accumulator)

        # close running generators
        running = False
        for i in range(len(tasks)):
            q.put_nowait(0)
        self.loop.run_until_complete(asyncio.wait(tasks, loop=self.loop))

        self.assertFalse(_q._sync_mutex.locked())
        _q.close()
        self.loop.run_until_complete(_q.wait_closed())

    def test_join_empty_queue(self):
        _q = self.q_class(loop=self.loop)
        q = _q.async_q

        # Test that a queue join()s successfully, and before anything else
        # (done twice for insurance).

        async def join():
            await q.join()
            await q.join()

        self.loop.run_until_complete(join())

        self.assertFalse(_q._sync_mutex.locked())
        _q.close()
        self.loop.run_until_complete(_q.wait_closed())

    @unittest.expectedFailure
    def test_format(self):
        _q = self.q_class(loop=self.loop)
        q = _q.async_q
        self.assertEqual(q._format(), 'maxsize=0')

        q._unfinished_tasks = 2
        self.assertEqual(q._format(), 'maxsize=0 tasks=2')

        self.assertFalse(_q._sync_mutex.locked())
        _q.close()
        self.loop.run_until_complete(_q.wait_closed())


class QueueJoinTests(_QueueJoinTestMixin, _QueueTestBase):
    q_class = janus.Queue


class LifoQueueJoinTests(_QueueJoinTestMixin, _QueueTestBase):
    q_class = janus.LifoQueue


class PriorityQueueJoinTests(_QueueJoinTestMixin, _QueueTestBase):
    q_class = janus.PriorityQueue
